;********************************************************************** 
; Spin a set of LEDs on a motor and switch them on and off such that
; they leave a message in mid-air.
;********************************************************************** 
;
; 03-Jan-2006
; Created PCB and template code.
; 
; 19-Oct-2008
; Wow. Haven't done anything on this in a while. Time to get it
; working, now I need a method for writing the letsmakerobots.com
; logo in the air.
;
; 23-Oct-2008
; Added an opto sensor which detects a black line as the device passes
; it. This "tells" the light writer when to start the message. Good
; opportunity to play with the comparator and Vref modules.
; 
; 24-Oct-2008
; Was getting appx 50Hz output from C1 comparator module. Grounded RA0
; so as not to cause comparator interrupts at this frequnecy. Didn't
; work, so it was tied to Vdd instead. The 50Hz has gone. The frequency
; may have been caused by the repeated test reading of CMCON.
;
; 25-Oct-2008
; Changed the execution so that the message display is the main
; program loop and is initiated by the interrupt. This makes the
; interrupt time shorter.
; Fixed an error in the lookup table for punctuation.
; Sorted out long addressed lookup tables. (Allowing more characters
; in the font.)
;
; 27-Oct-2008
; Now able to scroll a long message.
;
; 02-Nov-2008
; Calculates and displays rotation time.
; Used include file macros.
;
; 09-Nov-2008
; Remade the font, adding characters to make it a complete set of
; values from 32 to 126. Extensive modifications to code performing
; font lookup.
; Reworked the refresh so each successive column is output based on
; a timer interrupt, which is initiated by the comparator
; (opto-switch) interrupt.
; Removed the windowing code for long strings. buffering is best and
; it will allow dynamic strings.
;
; 04-Dec-2008
; Reworked the buffer. Was never really happy with it.
; Display a longer message in segments. Now much more stable.
;
;********************************************************************** 
;
; TODO
; Scroll message in from side.
; Other effects (eg scroll up from bottom).
;
;********************************************************************** 
;
; Written for a 16F628A @ 4MHz.

  processor 16F628A ; PIC type

; no code protect, !MCLR used for input, internal RC clock,
; no watchdog timer, powerup timer enabled, brownout detect enabled

  __config _CP_OFF & _LVP_OFF & _INTRC_OSC_NOCLKOUT & _MCLRE_OFF & _WDT_OFF & _PWRTE_ON & _BODEN_ON

  __idlocs 0x5D12

#include <p16f628a.inc>
#include <..\k1601f.inc>
#include <..\kmacro00a.inc>

; #define Debug -1

#IFDEF Debug
  MESSG "*** DEBUG BUILD !! ***"
  MESSG "*** DEBUG BUILD !! ***"
  MESSG "*** DEBUG BUILD !! ***"
#ENDIF

;**********************************************************************
; Useful Constants
;**********************************************************************

ChLen EQU 0x08 ; number of bytes per character
BufLen EQU .30 ; number of characters on display


;**********************************************************************
; I/O pins and registers
;**********************************************************************

PortOut   EQU PORTB  ; switch LEDs
TrisOut   EQU TRISB

;        EQU RB0    ; pin  6 ; 
;        EQU RB1    ; pin  7 ; 
;        EQU RB2    ; pin  8 ; 
;        EQU RB3    ; pin  9 ; 
;        EQU RB4    ; pin 10 ; 
;        EQU RB5    ; pin 11 ; 
;        EQU RB6    ; pin 12 ; 
;        EQU RB7    ; pin 13 ; 

Com      EQU RA1    ; pin 18 ; comparator input (opto sensor)

; ICSP connections are as follows:
; ICSP   PIC   func   colour
;   1     14   VDD     red
;   2     13   RB7/PGD wht
;   3     12   RB6/PGC blu
;   4      4   /MCLR   yel
;   5      5   VSS     bla


;**********************************************************************
; static RAM GPRs
;**********************************************************************

; define variables in static RAM (General Purpose Registers)

  CBLOCK Bank0GPR ; start of GPRs
    Ch                       ; ascii current character
    ChCol                    ; index within current character
    HexTmp                   ; temp byte in hex conversion
    T1old:2                  ; timer1 at start of refresh
    T1delta:2                ; change in T1 since last iteration
    DLC,d250ms, d1s          ; delay loop counters
  ENDC

  CBLOCK BankShare ; start of shared GPR RAM
    W_temp                   ; temporary copy of W register
    STATUS_temp              ; temporary copy of STATUS register
    FSR_temp                 ; temporary copy of FSR register
    Flags
    A:2                      ; handy for 16-bit arithmetic
    Mul:2, WA                ; use in muliplication
  ENDC

  CBLOCK Bank2GPR ; start of bank 2 GPRs
    BufPtrW, BufPtrR         ; buffer read and write pointers
    TempCh                   ; current character
    iCh                      ; address of char copying string to queue
    BufStart:BufLen          ; serial buffer
  ENDC


;**********************************************************************
; Flags bits
;**********************************************************************

FlgFlashBit  EQU 0  ; indicates if flashing LEDs are on or off
FlgBufEmpty  EQU 5  ; no more characters in buffer
FlgBufFull   EQU 6  ; buffer full


;**********************************************************************
; Character Constants
;**********************************************************************

ChNull    EQU 0x00 ; null character (signifies the end of a string)
ChCR      EQU "\r" ; 0x0D = carriage return (signifies the end of a string)
ChQuote   EQU 0x22 ; double quotes character (for inclusion in string)


;**********************************************************************
; Macro definitions
;**********************************************************************

M_CMPJLE MACRO Reg, Lit, Addr ; compare and jump if Register <= Literal 

  MOVF Reg, W 
  SUBLW Lit                  ; W = Lit - Reg 
  BTFSC STATUS, C            ; If !carry is clear, there was a carry, 
  GOTO Addr                  ; so Reg > Lit so skip the GOTO 

  ENDM ; M_CMPJLE 


; **********************************************************************

  org 0x0000                 ; start of program memory
	  GOTO Init


; **********************************************************************
; Interrupt vector
; **********************************************************************

  ORG 0x0004                 ; interrupt vector
  GOTO IntVector


; **********************************************************************
; String lookup tables
; **********************************************************************

LUTstring:
  CLRF PCLATH
  ADDWF PCL, F
StrALL  DT "All LMR Are Belong To Us!", ChCR ; 26
StrLETS DT "          Lets", ChCR
StrMAKE DT "          Make", ChCR
StrROBOTS DT "        Robots!", ChCR
StrLMR  DT " WWW.LETSMAKEROBOTS.COM", ChCR ; 23
StrMM   DT "   Matthew McMullan (P4)", ChCR ; 25
StrFI   DT "     Forge Integrated", ChCR ; 22
StrPS   DT "      Primary School", ChCR ; 21
; Str_ABC DT "ABCDEFGHIJKLMNOPQRSTUVWXYZ", ChCR
; Str_abc DT "abcdefghijklmnopqrstuvwxyz", ChCR
; Str_123 DT "0123456789 !.?", ChCR
; StrLong DT "THIS IS A TEST TO SEE IF WE CAN SCROLL A LONG MESSAGE...", ChCR


; **********************************************************************
; Other useful lookup tables
; **********************************************************************

LUTbitMask: ; convert bit index into bit mask
  ADDWF PCL, F
  DT 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80

LUThex: ; convert 4-bit value to hex ASCII equivalent
  ADDWF PCL, F
  DT "0123456789ABCDEF"


;**********************************************************************
; Font data
;**********************************************************************
  ORG 0x0100                 ; page 1 (base 0)

LUT_font:

  BANKSEL ChCol
; Ch      = ascii current character
; ChCol   = index within current character

  CLRF Mul+1
  MOVLW .32
  SUBWF Ch, W
  MOVWF Mul
  MOVLW 0x08
  CALL Mul16                 ; Mul = (Ch - 32) * 8

  MOVLW HIGH LUT_font
  MOVWF A+1
  MOVF ChCol, W              ; character column
  ADDWF PCL, W               ; add on (low) program counter
  ADDLW .8                   ; add on enough for the next 8 operations
  MOVWF A
  MOVLW Mul
  CALL Add16                 ; A = (PCL + 8 + Column):[LUT_font] + Mul
  
  MOVF A+1, W
  MOVWF PCLATH
  MOVF A, W
  MOVWF PCL

  DT 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ; <space>
  DT 0x00, 0x00, 0x00, 0x5F, 0x00, 0x00, 0x00, 0x00 ; !
  DT 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00 ; "
  DT 0x00, 0x24, 0x7E, 0x24, 0x24, 0x7E, 0x24, 0x00 ; #
  DT 0x00, 0x24, 0x2A, 0x7F, 0x2A, 0x12, 0x00, 0x00 ; $
  DT 0x00, 0x46, 0x26, 0x10, 0x08, 0x64, 0x62, 0x00 ; %
  DT 0x00, 0x32, 0x4D, 0x4D, 0x55, 0x22, 0x50, 0x00 ; &
  DT 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00 ; '
  DT 0x00, 0x00, 0x00, 0x1C, 0x22, 0x41, 0x00, 0x00 ; (
  DT 0x00, 0x00, 0x41, 0x22, 0x1C, 0x00, 0x00, 0x00 ; )
  DT 0x00, 0x22, 0x14, 0x7F, 0x14, 0x22, 0x00, 0x00 ; *
  DT 0x00, 0x08, 0x08, 0x3E, 0x08, 0x08, 0x00, 0x00 ; +
  DT 0x00, 0x00, 0x00, 0x80, 0x40, 0x00, 0x00, 0x00 ; ,
  DT 0x00, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, 0x00 ; -
  DT 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00 ; .
  DT 0x00, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x00 ; /

; 0..9

  DT 0x00, 0x3E, 0x45, 0x49, 0x51, 0x3E, 0x00, 0x00
  DT 0x00, 0x44, 0x42, 0x7F, 0x40, 0x40, 0x00, 0x00
  DT 0x00, 0x42, 0x61, 0x51, 0x49, 0x46, 0x00, 0x00
  DT 0x00, 0x22, 0x41, 0x49, 0x49, 0x36, 0x00, 0x00
  DT 0x00, 0x18, 0x14, 0x12, 0x7F, 0x10, 0x00, 0x00
  DT 0x00, 0x27, 0x45, 0x45, 0x45, 0x39, 0x00, 0x00
  DT 0x00, 0x3C, 0x4A, 0x49, 0x49, 0x30, 0x00, 0x00
  DT 0x00, 0x01, 0x01, 0x79, 0x05, 0x03, 0x00, 0x00
  DT 0x00, 0x36, 0x49, 0x49, 0x49, 0x36, 0x00, 0x00
  DT 0x00, 0x06, 0x49, 0x49, 0x29, 0x1E, 0x00, 0x00

  DT 0x00, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, 0x00 ; :
  DT 0x00, 0x00, 0x00, 0x80, 0x50, 0x00, 0x00, 0x00 ; ;
  DT 0x00, 0x08, 0x14, 0x14, 0x22, 0x22, 0x41, 0x00 ; <
  DT 0x00, 0x00, 0x28, 0x28, 0x28, 0x28, 0x00, 0x00 ; =
  DT 0x00, 0x41, 0x22, 0x22, 0x14, 0x14, 0x08, 0x00 ; >
  DT 0x00, 0x00, 0x02, 0x01, 0x59, 0x06, 0x00, 0x00 ; ?
  DT 0x1C, 0x22, 0x49, 0x15, 0x1D, 0x22, 0x1C, 0x00 ; @

; A..Z

  DT 0x00, 0x7C, 0x12, 0x11, 0x12, 0x7C, 0x00, 0x00
  DT 0x00, 0x7F, 0x49, 0x49, 0x49, 0x36, 0x00, 0x00
  DT 0x00, 0x3E, 0x41, 0x41, 0x41, 0x22, 0x00, 0x00
  DT 0x00, 0x7F, 0x41, 0x41, 0x22, 0x1C, 0x00, 0x00
  DT 0x00, 0x7F, 0x49, 0x49, 0x41, 0x41, 0x00, 0x00
  DT 0x00, 0x7F, 0x09, 0x09, 0x01, 0x01, 0x00, 0x00
  DT 0x00, 0x3E, 0x41, 0x41, 0x51, 0x72, 0x00, 0x00
  DT 0x00, 0x7F, 0x08, 0x08, 0x08, 0x7F, 0x00, 0x00
  DT 0x00, 0x00, 0x41, 0x7F, 0x41, 0x00, 0x00, 0x00
  DT 0x00, 0x21, 0x41, 0x3F, 0x01, 0x01, 0x00, 0x00
  DT 0x00, 0x7F, 0x08, 0x14, 0x22, 0x41, 0x00, 0x00
  DT 0x00, 0x7F, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00
  DT 0x00, 0x7F, 0x02, 0x04, 0x02, 0x7F, 0x00, 0x00
  DT 0x00, 0x7F, 0x04, 0x08, 0x10, 0x7F, 0x00, 0x00
  DT 0x00, 0x3E, 0x41, 0x41, 0x41, 0x3E, 0x00, 0x00
  DT 0x00, 0x7F, 0x09, 0x09, 0x09, 0x06, 0x00, 0x00
  DT 0x00, 0x3E, 0x41, 0x51, 0x21, 0x5E, 0x00, 0x00
  DT 0x00, 0x7F, 0x09, 0x19, 0x29, 0x46, 0x00, 0x00
  DT 0x00, 0x26, 0x49, 0x49, 0x49, 0x32, 0x00, 0x00
  DT 0x00, 0x01, 0x01, 0x7F, 0x01, 0x01, 0x00, 0x00
  DT 0x00, 0x3F, 0x40, 0x40, 0x40, 0x3F, 0x00, 0x00
  DT 0x00, 0x07, 0x18, 0x60, 0x18, 0x07, 0x00, 0x00
  DT 0x00, 0x3F, 0x40, 0x30, 0x40, 0x3F, 0x00, 0x00
  DT 0x00, 0x63, 0x14, 0x08, 0x14, 0x63, 0x00, 0x00
  DT 0x00, 0x03, 0x04, 0x78, 0x04, 0x03, 0x00, 0x00
  DT 0x00, 0x61, 0x51, 0x49, 0x45, 0x43, 0x00, 0x00

  DT 0x00, 0x00, 0x00, 0x7F, 0x41, 0x41, 0x00, 0x00 ; [
  DT 0x00, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x00 ; \
  DT 0x00, 0x00, 0x41, 0x41, 0x7F, 0x00, 0x00, 0x00 ; ]
  DT 0x00, 0x04, 0x02, 0x01, 0x02, 0x04, 0x00, 0x00 ; ^
  DT 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80 ; _
  DT 0x00, 0x00, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00 ; `

; a..z

  DT 0x00, 0x00, 0x20, 0x54, 0x54, 0x78, 0x00, 0x00
  DT 0x00, 0x00, 0x7E, 0x50, 0x50, 0x20, 0x00, 0x00
  DT 0x00, 0x00, 0x30, 0x48, 0x48, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x20, 0x50, 0x50, 0x7E, 0x00, 0x00
  DT 0x00, 0x00, 0x38, 0x54, 0x54, 0x18, 0x00, 0x00
  DT 0x00, 0x00, 0xF8, 0x14, 0x14, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x10, 0xA8, 0xA8, 0x78, 0x00, 0x00
  DT 0x00, 0x00, 0x7E, 0x10, 0x10, 0x60, 0x00, 0x00
  DT 0x00, 0x00, 0x00, 0x74, 0x00, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x80, 0x80, 0x74, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x7E, 0x20, 0x50, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x3E, 0x40, 0x40, 0x00, 0x00, 0x00
  DT 0x00, 0x70, 0x10, 0x20, 0x10, 0x60, 0x00, 0x00
  DT 0x00, 0x00, 0x70, 0x10, 0x10, 0x60, 0x00, 0x00
  DT 0x00, 0x00, 0x30, 0x48, 0x48, 0x30, 0x00, 0x00
  DT 0x00, 0x00, 0xF8, 0x28, 0x28, 0x10, 0x00, 0x00
  DT 0x00, 0x00, 0x10, 0x28, 0x28, 0xF8, 0x00, 0x00
  DT 0x00, 0x00, 0x78, 0x10, 0x08, 0x10, 0x00, 0x00
  DT 0x00, 0x00, 0x08, 0x54, 0x54, 0x20, 0x00, 0x00
  DT 0x00, 0x00, 0x3C, 0x50, 0x50, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x38, 0x40, 0x40, 0x38, 0x00, 0x00
  DT 0x00, 0x18, 0x20, 0x40, 0x20, 0x18, 0x00, 0x00
  DT 0x00, 0x38, 0x40, 0x20, 0x40, 0x38, 0x00, 0x00
  DT 0x00, 0x00, 0x48, 0x30, 0x30, 0x48, 0x00, 0x00
  DT 0x00, 0x00, 0x98, 0xA0, 0x78, 0x00, 0x00, 0x00
  DT 0x00, 0x00, 0x48, 0x68, 0x58, 0x48, 0x00, 0x00

  DT 0x00, 0x00, 0x08, 0x36, 0x41, 0x41, 0x00, 0x00 ; {
  DT 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00 ; |
  DT 0x00, 0x41, 0x41, 0x36, 0x08, 0x00, 0x00, 0x00 ; }
  DT 0x04, 0x02, 0x02, 0x04, 0x08, 0x08, 0x04, 0x00 ; ~
  

;**********************************************************************
  ORG 0x0500                 ; page 5 (base 0)

;********************************************************************** 
; New Interrupt Vector 
;********************************************************************** 

IntVector: 

; Interrupts are serviced here instead of at 0x0004 as this permits the 
; creation of the lookup tables before the end of the first memory page. 

; save context 

  ERRORLEVEL -302 

  MOVWF W_temp               ; Copy W to TEMP register, 
  SWAPF STATUS, W            ; Swap status to be saved into W 
  MOVWF STATUS_temp          ; Save status to STATUStemp register 
  MOVF FSR, W                ; temp copy of FSR 
  MOVWF FSR_temp 

  ERRORLEVEL +302 

  BANKSEL PORTA              ; bank 0

  BTFSC PIR1, CMIF           ; was it a comparator interrupt?
  GOTO CMP_Handler           ; yes

  BTFSC INTCON, T0IF         ; was it a timer 0 overflow?
  GOTO Tmr0_Handler

; otherwise

  GOTO DoneInt

Tmr0_Handler:

  CALL NextCol
  BCF INTCON, T0IF           ; clear timer 0 interrupt
  GOTO DoneInt

CMP_Handler:
  BTFSS CMCON, C2OUT         ; bit 7 set => white, so start message
  GOTO DoneCMP               ; otherwise
  BSF INTCON, T0IE           ; start message by allowing T0 interrupts

; calculate the time since last start (full rotation time)

  M_LogTimer1 A              ; log timer 1 to A and take a 
  M_Move16 A, T1delta        ; temporary copy to T1delta
  
  MOVLW T1old
  CALL Sub16                 ; A = A - T1old

  M_Move16 T1delta, T1old    ; keep historic note of timer 1
  
  M_Move16 A, T1delta
  
DoneCMP:
  BCF PIR1, CMIF             ; allow subsequent comparator interrupts
  
DoneInt: 

  ; restore context 

  ERRORLEVEL -302 

  MOVF FSR_temp, W           ; restore FSR 
  MOVWF FSR 
  SWAPF STATUS_temp, W       ; swap nibbles in STATUStemp, result in W 
  MOVWF STATUS               ; W into STATUS register (sets bank to original state) 
  SWAPF W_temp, F            ; swap nibbles in Wtemp 
  SWAPF W_temp, W            ; swap nibbles in Wtemp, result into W 

  ERRORLEVEL +302 

  RETFIE                     ; no - end ISR 


; ********************************************************************** 

NextCol: ; index the column and display it

  BANKSEL ChCol

  INCF ChCol, F              ; increment column

TestChCol: ; check for column overflow
  M_CMPJNE ChCol, ChLen, NoResetChCol ; ChCol <> ChLen so do nothing
  CLRF ChCol ; start the next character at the first row index

  CALL Dequeue               ; dequeue next character
  MOVWF Ch

NoResetChCol:  
  M_CMPJNE Ch, ChCR, NoResetMsg ; char <> CR, no reset
  CALL ResetReadPtr
  BCF INTCON, T0IE           ; end message by disallowing T0 interrupts
  BCF Flags, FlgBufEmpty     ; buffer isn't really "empty"
  GOTO DoneNextCol

NoResetMsg:
  CALL LUT_font
  MOVWF PortOut

DoneNextCol:
  RETURN ; NextCol


; **********************************************************************

ResetReadPtr: ; reset message buffer read pointer

  BANKSEL ChCol
  MOVLW ChLen -1
  MOVWF ChCol

  BANKSEL BufStart
  MOVLW LOW BufStart         ; reset queue read pointer
  MOVWF low BufPtrR

  RETURN ; ResetReadPtr
  

; **********************************************************************

ResetWritePtr: ; reset message buffer write pointer

  BANKSEL BufStart
  MOVLW LOW BufStart         ; reset queue write pointer
  MOVWF low BufPtrW

  BCF Flags, FlgBufEmpty
  BCF Flags, FlgBufFull
  
  RETURN ; ResetWritePtr
  

; ********************************************************************** 

Dequeue: ; read next character from output queue into W

  ERRORLEVEL -302 
  BANKSEL BufStart 

  BTFSC Flags, FlgBufEmpty 
  GOTO DoneDQ 

  MOVF low BufPtrR, W       ; copy pointer to file select register 
  MOVWF FSR 

; update read pointer
  MOVLW LOW BufStart + BufLen - 1 ; end of buffer 
  XORWF low BufPtrR, W       ; reached? 
  BTFSS STATUS, Z
  GOTO BufIncR               ; no: increment buffer read pointer 
  BSF Flags, FlgBufEmpty     ; yes: buffer empty 
  MOVLW 0x00
  GOTO DoneDQ

BufIncR:
  INCF low BufPtrR, F
  BANKISEL BufStart 
  MOVF INDF, W               ; get buffer 
  
  ERRORLEVEL +302 

DoneDQ: 
  BANKSEL PORTA 
  BANKISEL PORTA 
  RETURN ; Dequeue 


; ********************************************************************** 

Enqueue: ; push character in W onto output queue 

  ERRORLEVEL -302 
  BANKSEL BufStart 
  BANKISEL BufStart 

  BTFSC Flags, FlgBufFull    ; if the buffer is full, ignore the byte 
  GOTO DoneNQ

  MOVWF low TempCh           ; temporary copy 

  MOVF low BufPtrW, W        ; copy pointer to file select register 
  MOVWF FSR 
  MOVF low TempCh, W         ; copy character to queue 

  BANKISEL BufStart 
  MOVWF INDF 
  BANKISEL PORTA 

; update write pointer
  MOVLW LOW BufStart + BufLen - 1 ; end of buffer 
  XORWF LOW BufPtrW, W       ; reached?
  BTFSS STATUS, Z
  GOTO BufIncW               ; no: increment buffer write pointer 
  BSF Flags, FlgBufFull      ; yes: buffer full 
  GOTO DoneNQ

BufIncW:  
  INCF low BufPtrW, F

DoneNQ:
  ERRORLEVEL +302 
  
  BANKSEL PORTA              ; the bank most used by the main program 
  RETURN ; Enqueue 


; **********************************************************************

HexOut: ; appends the byte in W onto the output queue

; Passes the MS nibble of the value to LUThex to obtain the ASCII code
; for the hex value of the nibble and outputs this as the most
; significant hex digit. Next, the LS nibble of the value is passed to
; LUThex to obtain the ASCII code for the hex value of the nibble and
; this is output as the most significant hex digit.
 
  MOVF HIGH LUThex, W
  MOVWF PCLATH

  BANKSEL HexTmp
  MOVWF HexTmp               ; copy W

  SWAPF HexTmp, W            ; swap original into W
  ANDLW 0x0F                 ; mask out MS nibble
  CALL LUThex
  CALL Enqueue

  BANKSEL HexTmp
  MOVF HexTmp, W             ; copy original into W
  ANDLW 0x0F                 ; mask out MS nibble
  CALL LUThex
  CALL Enqueue

  RETURN ; HexOut


;**********************************************************************
; Adds the 16-bit word whose low byte is at W to the 16 bit
; word AL:AH and leaves the result in AL:AH.

Add16

  MOVWF FSR                  ; address of low byte of word to be added to A
  
  MOVF INDF, W               ; first number low byte
  ADDWF A, F                 ; second number low byte

  INCF FSR, F                ; go to first number high byte
  MOVF INDF, W               ; first number high byte
  BTFSC STATUS, C            ; skip next if C is not set
  INCFSZ INDF, W             ; increment first number high byte
  ADDWF A+1, F               ; second number high byte

  RETURN ; Add16


;**********************************************************************
; Subtracts the 16-bit word whose low byte is at W from the 16 bit
; word AL:AH and leaves the result in AL:AH.

Sub16

  MOVWF FSR                  ; address of low byte of word to be added to A
  
  MOVF INDF, W               ; first number low byte
  SUBWF A, F                 ; second number low byte

  INCF FSR, F                ; go to first number high byte
  MOVF INDF, W               ; first number high byte
  BTFSS STATUS, C            ; skip next if C is set (ie result is 0 or +ve)
  INCFSZ INDF, W             ; increment first number high byte
  SUBWF A+1, F               ; second number high byte

  RETURN ; Sub16


;**********************************************************************
; Multiplies the 8-bit number in W by the 16-bit value Mul:MulH.
; The result is left in Mul. Interferes with A and WA.

Mul16: ; Mul:MulH = Mul:MulH * W

  MOVWF WA          ; Temp copy of W

  MOVF Mul, W       ; check Mul != 0
  IORWF Mul+1, W
  BTFSC STATUS, Z
  RETURN            ; exit with 0 in Mul:MulH

  MOVF WA, W        ; check Temp != 0
  BTFSS STATUS, Z
  GOTO Mul16Start
  CLRF Mul          ; clear Mul:MulH
  CLRF Mul+1
  RETURN            ; exit with 0 in Mul:MulH

Mul16Start: 
  CLRF A            ; clear AL:AH
  CLRF A+1

LoopMul16:
  MOVLW Mul         ; repeatedly add Mul to AL:AH
  CALL Add16
  DECFSZ WA, F      ; decrementing WA
  GOTO LoopMul16    ; until WA = 0

  MOVF A, W         ; copy AL:AH into Mul:MulH
  MOVWF Mul
  MOVF A+1, W
  MOVWF Mul+1

  RETURN ; Mul16


; **********************************************************************

EnqueueStr: ; push the string whose address is in W onto output queue

  ERRORLEVEL -302
  BANKSEL BufStart

  BTFSC Flags, FlgBufFull    ; if the buffer is full, ignore the sequence
  GOTO DoneNQstr
  
  MOVWF iCh                  ; location of the start of the string
  MOVLW LUTstring + 2        ; subtract location of LUTstring
  SUBWF iCh, F

NQloop:
  MOVF iCh, W                ; get next character from string
  CALL LUTstring
  CALL Enqueue               ; push character in W onto output queue

  BANKSEL BufStart
  MOVF iCh, W                ; get next character from string
  CALL LUTstring
  XORLW ChCR                 ; check for end of string
  BTFSC STATUS, Z
  GOTO DoneNQ                ; terminate if end of string

  INCF iCh, F
  GOTO NQloop

  ERRORLEVEL +302

DoneNQstr:
  BANKSEL PORTA              ; the bank most used by the main program
  RETURN ; EnqueueStr


; **********************************************************************

Delay1ms ; 1ms delay (248 cycles of NOP, DECFSZ, GOTO = 991 cycles)
  MOVLW 0xF8     ; +1 + 2 for call = 3
  MOVWF DLC      ; +1 = 4
D1Loop
  NOP            ; +1
  DECFSZ DLC, F  ; +1 (+1 for last iteration)
  GOTO D1Loop    ; +2 (-2 for last iteration) = 995 total
  NOP            ; +1 = 996
  NOP            ; +1 = 997
  NOP            ; +1 = 998
  RETURN         ; +2 for return = 1000


Delay250ms:
  MOVLW d'250'
  MOVWF d250ms
D250loop:
  CALL Delay1ms
  DECFSZ d250ms, F
  GOTO D250loop

  RETURN

Delay1s:
  BANKSEL d1s
  MOVLW d'4'
  MOVWF d1s
D1sloop:
  CALL Delay250ms
  DECFSZ d1s, F
  GOTO D1sloop

  RETURN


;**********************************************************************

Init:

  BANKSEL PORTA              ; (set bank 0)

  BCF PIR1, CMIF             ; Clear pending comparator interrupts
  BCF INTCON, T0IF           ; clear pending timer 0 interrupts
  BCF INTCON, T0IE           ; mask timer 0 interrupts
  BSF INTCON, PEIE           ; Enable peripheral interrupts

; clear ports

  CLRF PORTA                 ; clear ports
  CLRF PORTB

; C1 is RA0 not RA3, C2 is RA1 not RA2
; bit 7 = (C2Vin+ > C2Vin-), ie Vref > RA1
; bit 6 = (C1Vin+ > C1Vin-), ie Vref > RA0

  MOVLW b'00000010'          ; RA0 and RA1 connected to C1 and C2
  MOVWF CMCON

  BANKSEL TRISA              ; (set bank 1)
  ERRORLEVEL -302	

  BSF PIE1, CMIE             ; Enable comparator interrupts
  
; set up I/O

  MOVLW b'10101010'          ; Vref = (10 / 24) * 6 = 2.5V
  MOVWF VRCON

  MOVLW 0xFF
  MOVWF TRISA                ; as input
  CLRF TrisOut               ; as output
  
  MOVF OPTION_REG, W         ; timer 0 1:4 prescaler
  ANDLW 0xD0
  IORLW b'00000001'
  MOVWF OPTION_REG

  BANKSEL T1CON
  BSF T1CON, TMR1ON          ; start timer 
  BSF T1CON, T1CKPS0         ; 1:8 prescaler
  BSF T1CON, T1CKPS1

  ERRORLEVEL +302


; **********************************************************************

Main:

  CLRF Flags
  BSF Flags, FlgBufEmpty

  ERRORLEVEL +302 

  CALL ResetReadPtr

  CALL ResetWritePtr
  MOVLW LOW StrALL
  CALL EnqueueStr

  BSF INTCON, GIE            ; Global interrupt enable

Loop:

  CALL Delay1s
  CALL Delay1s
  CALL Delay1s

  CALL ResetWritePtr
  MOVLW LOW StrLETS
  CALL EnqueueStr

  CALL Delay1s

  CALL ResetWritePtr
  MOVLW LOW StrMAKE
  CALL EnqueueStr

  CALL Delay1s

  CALL ResetWritePtr
  MOVLW LOW StrROBOTS
  CALL EnqueueStr

  CALL Delay1s

  CALL ResetWritePtr
  MOVLW LOW StrLMR
  CALL EnqueueStr

  CALL Delay1s
  CALL Delay1s
  CALL Delay1s

  CALL ResetWritePtr
  MOVLW LOW StrALL
  CALL EnqueueStr

  GOTO Loop

  END
